"""Parser for FusionCatcher"""
# Copyright (c) 2015-2022 Wibowo Arindrarto <contact@arindrarto.dev>
# SPDX-License-Identifier: BSD-3-Clause

from os import PathLike
from typing import Dict, List, TextIO, Union

import click

from .utils import get_handle

__all__ = ["parse"]

# Expected column names
_COLS = {
    "0.99.5a": [
        "Gene_1_symbol(5end_fusion_partner)",
        "Gene_2_symbol(3end_fusion_partner)",
        "Fusion_description",
        "Counts_of_common_mapping_reads",
        "Spanning_pairs",
        "Spanning_unique_reads",
        "Longest_anchor_found",
        "Fusion_finding_method",
        "Fusion_point_for_gene_1(5end_fusion_partner)",
        "Fusion_point_for_gene_2(3end_fusion_partner)",
        "Gene_1_id(5end_fusion_partner)",
        "Gene_2_id(3end_fusion_partner)",
        "Exon_1_id(5end_fusion_partner)",
        "Exon_2_id(3end_fusion_partner)",
        "Fusion_sequence",
        "Predicted_effect",
        "Predicted_fused_transcripts",
        "Predicted_fused_proteins",
    ],
    "1.00": [
        "Gene_1_symbol(5end_fusion_partner)",
        "Gene_2_symbol(3end_fusion_partner)",
        "Fusion_description",
        "Counts_of_common_mapping_reads",
        "Spanning_pairs",
        "Spanning_unique_reads",
        "Longest_anchor_found",
        "Fusion_finding_method",
        "Fusion_point_for_gene_1(5end_fusion_partner)",
        "Fusion_point_for_gene_2(3end_fusion_partner)",
        "Gene_1_id(5end_fusion_partner)",
        "Gene_2_id(3end_fusion_partner)",
        "Exon_1_id(5end_fusion_partner)",
        "Exon_2_id(3end_fusion_partner)",
        "Fusion_sequence",
        "Predicted_effect",
    ],
    # The non empty output of version 1.20 is the same as version 1.00
    "1.20-empty": [
        "Fusion_gene_symbol_1(5end_partner)",
        "Fusion_gene_symbol_2(3end_partner)",
        "Fusion_description",
        "Counts_of_common_mapping_reads",
        "Spanning_pairs",
        "Spanning_unique_reads",
        "Longest_spanning_read",
        "Aligner(s)",
        "Fusion_gene_1_position(5end_partner)",
        "Fusion_gene_2_position(3end_partner)",
        "Fusion_gene_1(5end_partner)",
        "Fusion_gene_2(3end_partner)",
        "Fusion_exon_1(5end_partner)",
        "Fusion_exon_2(3end_partner)",
        "Fusion_sequence",
    ],
}

# Delimiter strings
_DELIM: Dict[str, str] = {
    # Fusion description
    "desc": ",",
    # Chromosome-coordinate-strand
    "loc": ":",
    # Others
    "gen": ";",
}


def parse_lr_entry(
    lr_gene: str,
    lr_brkpoint: str,
) -> Dict[str, Union[str, int]]:
    """Parse the gene and breakpoint entry.

    :param lr_gene: Column value for 'LeftGene' or 'RightGene'.
    :param lr_brkpoint: Column value for 'LeftBreakpoint' or 'RightBreakpoint'.

    """
    lrgname, lrgid = lr_gene.split(_DELIM["gids"])
    lrchrom, lrpos, lrstrand = lr_brkpoint.split(_DELIM["loc"])

    return {
        "geneName": lrgname,
        "geneID": lrgid,
        "chromosome": lrchrom,
        "position": int(lrpos),
        "strand": lrstrand,
    }


def split_filter(string: str, delim: str) -> List[str]:
    """Split the given string with the given delimiter.

    If the given string is empty, an empty list is returned.

    :param string: String to split.
    :param delim: Delimiter character.

    """
    if not string:
        return []
    return string.split(delim)


def parse_raw_line(raw_line: str, colnames: List[str]) -> dict:
    """Parse a single line into a dictionary.

    :param raw_line: FusionCatcher result line.
    :param colnames: Column names present in the file.

    """
    values = raw_line.split("\t")
    if len(values) != len(colnames):
        msg = "Line values {0} does not match column names {1}."
        raise click.BadParameter(msg.format(values, colnames))

    d = {k: v.strip() for k, v in zip(colnames, values)}

    f5 = d["Fusion_point_for_gene_1(5end_fusion_partner)"].split(_DELIM["loc"])
    f3 = d["Fusion_point_for_gene_2(3end_fusion_partner)"].split(_DELIM["loc"])

    res = {
        "5end": {
            "geneSymbol": d["Gene_1_symbol(5end_fusion_partner)"],
            "geneID": d["Gene_1_id(5end_fusion_partner)"],
            "exonID": d["Exon_1_id(5end_fusion_partner)"] or None,
            "chromosome": f5[0],
            "position": int(f5[1]),
            "strand": f5[2],
        },
        "3end": {
            "geneSymbol": d["Gene_2_symbol(3end_fusion_partner)"],
            "geneID": d["Gene_2_id(3end_fusion_partner)"],
            "exonID": d["Exon_2_id(3end_fusion_partner)"] or None,
            "chromosome": f3[0],
            "position": int(f3[1]),
            "strand": f3[2],
        },
        "fusionDescription": split_filter(d["Fusion_description"], _DELIM["desc"]),
        "nCommonMappingReads": int(d["Counts_of_common_mapping_reads"]),
        "nSpanningPairs": int(d["Spanning_pairs"]),
        "nSpanningUniqueReads": int(d["Spanning_unique_reads"]),
        "longestAnchorLength": int(d["Longest_anchor_found"]),
        "fusionFindingMethod": split_filter(d["Fusion_finding_method"], _DELIM["gen"]),
        "fusionSequence": d["Fusion_sequence"],
        "predictedEffect": d["Predicted_effect"],
    }

    # Does predicted_fused_transcripts exist for the column format
    if "Predicted_fused_transcripts" in d:
        FusedTranscripts = split_filter(d["Predicted_fused_transcripts"], _DELIM["gen"])
        res["predictedFusedTranscripts"] = FusedTranscripts

    if "Predicted_fused_proteins" in d:
        FusedProteins = split_filter(d["Predicted_fused_proteins"], _DELIM["gen"])
        res["predictedFusedProteins"] = FusedProteins

    return res


def parse(in_data: Union[str, PathLike, TextIO]) -> List[dict]:
    """Parse the abridged output of a FusionCatcher run.

    :param in_data: Input FusionCatcher contents.
    :returns: Parsed values.

    """
    payload = []
    with get_handle(in_data) as src:
        first_line = src.readline().strip()
        # Parse column names
        colnames = first_line.split("\t")

        # If none of the colnames are recognised
        if not any(colnames == cols for cols in _COLS.values()):
            msg = "Unexpected column names: {0}."
            raise click.BadParameter(msg.format(colnames))

        for line in src:
            parsed = parse_raw_line(line, colnames)
            payload.append(parsed)

    return payload



"""Main entry point for command line invocation"""
# Copyright (c) 2015-2022 Wibowo Arindrarto <contact@arindrarto.dev>
# SPDX-License-Identifier: BSD-3-Clause

from pathlib import Path
from typing import Optional, TextIO, cast

import click

from . import __version__
from . import fastqc as m_fastqc
from . import flagstat as m_flagstat
from . import fusioncatcher as m_fusioncatcher
from . import picard as m_picard
from . import star as m_star
from . import star_fusion as m_star_fusion
from . import vep as m_vep
from .utils import write_output


@click.group()
@click.version_option(__version__, message="%(version)s")
@click.option(
    "--fmt",
    default="json",
    type=click.Choice(["json", "yaml"]),
    help="Output file format. Default: json.",
)
@click.option(
    "--indent",
    default=2,
    help="Indentation level. Ignored if the --compact flag is set. Default: 2.",
)
@click.option(
    "--compact",
    is_flag=True,
    help="Whether to create a compact JSON or not. Ignored if output format is"
    " YAML.",
)
@click.pass_context
def main(ctx: click.Context, fmt: str, indent: int, compact: bool) -> None:
    """Converts bioinformatics tools' output to a standard format."""
    ctx.params["fmt"] = fmt
    ctx.params["indent"] = indent
    ctx.params["compact"] = compact


@main.command()
@click.argument("input", type=click.Path(exists=True, path_type=str))
@click.argument("output", type=click.File("w"), default="-")
@click.pass_context
def fastqc(ctx: click.Context, input: str, output: TextIO) -> None:
    """Converts FastQC output.

    Use "-" for stdin and/or stdout.

    """
    payload = m_fastqc.parse(Path(input))
    parent = cast(click.Context, ctx.parent)
    write_output(payload, output, **parent.params)


@main.command()
@click.argument("input", type=click.File("r"))
@click.argument("output", type=click.File("w"), default="-")
@click.pass_context
def flagstat(ctx: click.Context, input: TextIO, output: TextIO) -> None:
    """Converts samtools flagstat output.

    Use "-" for stdin and/or stdout.

    """
    payload = m_flagstat.parse(input)
    parent = cast(click.Context, ctx.parent)
    write_output(payload, output, **parent.params)


@main.command()
@click.argument("input", type=click.File("r"))
@click.argument("output", type=click.File("w"), default="-")
@click.pass_context
def fusioncatcher(ctx: click.Context, input: TextIO, output: TextIO) -> None:
    """Converts FusionCatcher output.

    Use "-" for stdin and/or stdout.

    """
    payload = m_fusioncatcher.parse(input)
    parent = cast(click.Context, ctx.parent)
    write_output(payload, output, **parent.params)


@main.command()
@click.argument("input", type=click.File("r"))
@click.argument("output", type=click.File("w"), default="-")
@click.option(
    "--input-linesep",
    default=None,
    type=click.Choice(["posix", "windows"]),
    help=(
        "Line separator for input files; used when parsing. Default: native"
        " value for current operating system."
    ),
)
@click.pass_context
def picard(
    ctx: click.Context,
    input: TextIO,
    output: TextIO,
    input_linesep: Optional[str],
) -> None:
    """Converts Picard metrics output.

    Use "-" for stdin and/or stdout.

    """
    payload = m_picard.parse(input, input_linesep)
    parent = cast(click.Context, ctx.parent)
    write_output(payload, output, **parent.params)


@main.command()
@click.argument("input", type=click.File("r"))
@click.argument("output", type=click.File("w"), default="-")
@click.option(
    "--input-linesep",
    default=None,
    type=click.Choice(["nt", "posix"]),
    help=(
        "Line separator for input files; used when parsing. Default: native"
        " value for current operating system."
    ),
)
@click.pass_context
def star(
    ctx: click.Context,
    input: TextIO,
    output: TextIO,
    input_linesep: Optional[str],
) -> None:
    """Converts STAR Log.final.out file.

    Use "-" for stdin and/or stdout.

    """
    payload = m_star.parse(input, input_linesep)
    parent = cast(click.Context, ctx.parent)
    write_output(payload, output, **parent.params)


@main.command(name="star-fusion")
@click.argument("input", type=click.File("r"))
@click.argument("output", type=click.File("w"), default="-")
@click.pass_context
def star_fusion(ctx: click.Context, input: TextIO, output: TextIO) -> None:
    """Converts output of STAR-Fusion.

    Use "-" for stdin and/or stdout.

    """
    payload = m_star_fusion.parse(input)
    parent = cast(click.Context, ctx.parent)
    write_output(payload, output, **parent.params)


@main.command(name="vep")
@click.argument("input", type=click.File("r"))
@click.argument("output", type=click.File("w"), default="-")
@click.option(
    "--input-linesep",
    default=None,
    type=click.Choice(["nt", "posix"]),
    help=(
        "Line separator for input files; used when parsing. Default: native"
        " value for current operating system."
    ),
)
@click.pass_context
def vep(
    ctx: click.Context,
    input: TextIO,
    output: TextIO,
    input_linesep: Optional[str],
) -> None:
    """Converts plain text output of Variant Effect Predictor.

    Use "-" for stdin and/or stdout.

    """
    payload = m_vep.parse(input)
    parent = cast(click.Context, ctx.parent)
    write_output(payload, output, **parent.params)



"""General utilities"""
# Copyright (c) 2015-2022 Wibowo Arindrarto <contact@arindrarto.dev>
# SPDX-License-Identifier: BSD-3-Clause

import json
import os
import re
from contextlib import contextmanager
from os import PathLike, linesep
from pathlib import Path
from typing import cast, IO, Generator, List, Optional, TextIO, Union

import click
import yaml

LINESEPS = {"nt": "\r\n", "posix": "\n"}
RE_INT = re.compile(r"^([-+]?\d+)L?$")
RE_FLOAT = re.compile(r"^([-+]?\d*\.?\d+(?:[eE][-+]?[0-9]+)?)$")


def convert(raw_str: str) -> Union[str, int, float]:
    """Tries to convert a string to an int, float, or return it unchanged.

    The function tries first to convert to an int. If it fails, it tries
    to convert to a float. If it fails, the origin input string is returned.

    :param raw_str: Input string.

    """
    maybe_int = RE_INT.search(raw_str)
    if maybe_int is not None:
        return int(maybe_int.group(1))

    maybe_float = RE_FLOAT.search(raw_str)
    if maybe_float is not None:
        return float(maybe_float.group(1))

    return raw_str


def write_output(
    payload: Union[dict, List[dict]],
    out_handle: TextIO,
    fmt: str = "json",
    compact: bool = False,
    indent: int = 4,
) -> None:
    """Writes the given dictionary as JSON or YAML to the output handle.

    :param payload: Payload to write.
    :param out_handle: Output handle.
    :param fmt: Output format.
    :param compact: Whether to write a compact JSON or not. Ignored if the
        output format is JSON.
    :param indent: Indentation level (ignored if output ``compact`` is true).

    """
    if fmt == "json":
        if compact:
            json.dump(
                payload, out_handle, sort_keys=True, indent=None, separators=(",", ":")
            )
        else:
            json.dump(payload, out_handle, sort_keys=True, indent=indent)
            out_handle.write(linesep)
    else:
        out_handle.write(yaml.dump(payload, default_flow_style=False, indent=indent))


@contextmanager
def get_handle(
    input: Union[str, PathLike, IO],
    encoding: Optional[str] = None,
    mode: str = "r",
) -> Generator[TextIO, None, None]:
    """Context manager for opening files.

    This function returns a file handle of the given file name. You may also
    give an open file handle, in which case the file handle will be yielded
    immediately. The purpose of this is to allow the context manager handle
    both file objects and file names as inputs.

    If a file handle is given, it is not closed upon exiting the context.
    If a file name is given, it will be closed upon exit.

    :param input: Handle of open file or file name.
    :param encoding: Encoding of the file. Ignored if input is file handle.
    :param mode: Mode for opening file. Ignored if input is file handle.

    """
    if isinstance(input, (str, Path)):
        fh = cast(TextIO, click.open_file(f"{input}", mode=mode, encoding=encoding))
    else:
        fh = cast(TextIO, input)

    yield fh

    if isinstance(input, str):
        fh.close()


def get_linesep(system: Optional[str]) -> str:
    """Returns the line separator character(s) for the given system"""

    if system is None:
        return os.linesep

    linesep = LINESEPS.get(system.lower())
    if linesep is not None:
        return linesep

    raise ValueError(f"Can not resolve linesep for system {system!r}")



"""Parser for VEP plain text statistics file"""
# Copyright (c) 2015-2022 Wibowo Arindrarto <contact@arindrarto.dev>
# SPDX-License-Identifier: BSD-3-Clause

import collections
from os import PathLike
from typing import Dict, List, Optional, TextIO, Tuple, Union

import click

from .utils import convert, get_handle, get_linesep

__all__ = ["parse"]


# Default maximum file size ~ because our parser is not a streaming parser.
_MAX_SIZE = 1024 * 500  # 500 Kb


def parse_raw_value(raw_value: str, linesep: str) -> List[List[str]]:
    """Parse raw values from VEP"""
    parsed = list()
    for line in raw_value.strip().split(linesep):
        # This is the regular case
        if "\t" in line:
            parsed.append(line.split("\t", 1))
        # Special cases where VEP was run on an emtpy input file
        elif line == "Lines of input read":
            parsed.append(["Lines of input read", "0"])
        elif line == "Variants processed":
            parsed.append(["Variants processed", "0"])
    return parsed


def group2entry(
    group: str, linesep: str
) -> Tuple[str, Union[List[Union[str, int, float]], Dict[str, Union[str, int, float]]]]:
    """Given the raw string of a VEP statistics group, parse it into a
    key, value tuple.

    Basically, turning this:

        [Variant classes]
        deletion    18
        insertion   35
        SNV 448

    into this:

        ("Variant classes", {
            "deletion": 18,
            "insertion": 35,
            "SNV": 448
         })

    except for then the VEP group represents a histogram, which is the
    case for groups whose name starts with "Distribution of variants ...".

    Those groups, which may look like this:

        [Distribution of variants on chromosome 1]
        0   1
        1   40
        2   35
        3   50
        ...

    gets turned into this:

        ("Distribution of variants on chromosome 1",
         [1, 40, 35, 50, ...])

    """
    # If there are no values for this section, return a default dict of
    # integers, so that any property a user requests will be 0
    if linesep not in group:
        # Remove the brackets from the key
        key = group[1:-1]
        return key, collections.defaultdict(int)

    raw_key, raw_value = group.split(linesep, 1)

    # Remove the brackets from the key
    key = raw_key[1:-1]

    values = parse_raw_value(raw_value, linesep)

    if not key.startswith("Distribution of variants on"):
        valued = {k: convert(v) for k, v in values}
        return key, valued

    return key, [convert(v) for _, v in values]


def parse(
    in_data: Union[str, PathLike, TextIO],
    input_linesep: Optional[str] = None,
    max_size: int = _MAX_SIZE,
) -> dict:
    """Parse a VEP plain text statistics file into a dictionary.

    :param in_data: Input VEP statistics contents.
    :param input_linesep: Name of the operating system used for determining
        input line separator. Valid values are 'nt', 'posix', or None.
    :param max_size: Maximum expected size of input contents (default: 500 KiB).

    """
    with get_handle(in_data) as fh:
        contents = fh.read(max_size)

    if not contents.startswith("[VEP run statistics]"):
        msg = "Unexpected file structure. No contents parsed."
        raise click.BadParameter(msg)

    linesep = get_linesep(input_linesep)
    entries = [group2entry(g, linesep) for g in contents.split(linesep * 2)]

    return dict(entries)



"""Bioinformatics tool outputs converter to JSON or YAML"""
# Copyright (c) 2015-2022 Wibowo Arindrarto <contact@arindrarto.dev>
# SPDX-License-Identifier: BSD-3-Clause

from pathlib import Path

from single_source import get_version


__author__ = "Wibowo Arindrarto"
__contact__ = "contact@arindrarto.dev"
__homepage__ = "https://github.com/bow/crimson"
__version__ = get_version(__name__, Path(__file__).parent.parent)

del Path, get_version



"""Parser for FastQC output"""
# Copyright (c) 2015-2022 Wibowo Arindrarto <contact@arindrarto.dev>
# SPDX-License-Identifier: BSD-3-Clause

from io import StringIO
from os import PathLike, walk
from pathlib import Path
from typing import IO, Any, Dict, List, TextIO, Union, cast
from zipfile import ZipFile, is_zipfile

import click

from .utils import convert, get_handle

__all__ = ["parse"]

_MAX_SIZE = 1024 * 1024 * 10
_MAX_LINE_SIZE = 1024
_RESULTS_FNAME = "fastqc_data.txt"

# A module content can be a dictionary (when the module is 'Basic Statistics')
# or a list of dictionaries (keyed by the column name).
FastQCModuleContents = Union[Dict[str, Any], List[Dict[str, Any]]]

FastQCModulePayload = Dict[str, Union[str, FastQCModuleContents]]


class FastQCModule:

    """Class representing a FastQC analysis module."""

    def __init__(self, raw_lines: List[str], end_mark: str = ">>END_MODULE") -> None:
        """Initialize an instance.

        :param raw_lines: List of lines in the module.
        :param end_mark: Mark of the end of the module.

        """
        # The values of these will be set by _parse.
        self.status: str
        self.name: str

        self.extra: Dict[str, Any] = {}

        self.raw_lines = raw_lines
        self.end_mark = end_mark
        self.contents = self._parse()

    @property
    def dict(self) -> FastQCModulePayload:
        """Module data as a dictionary."""
        return {
            "contents": self.contents,
            "status": self.status,
            **self.extra,
        }

    def _parse(self) -> FastQCModuleContents:
        """Common parser for a FastQC module."""

        # Helper function for converting FastQC values that keeps
        # the "Base" column as strings (since it can be a number
        # or a strin denoting a range)
        def fqc_convert(k: str, v: Any) -> Any:
            if k == "Base":
                return v
            return convert(v)

        # check that the last line is a proper end mark
        if not self.raw_lines[-1].startswith(self.end_mark):
            raise ValueError(
                "Module last line does not start with the expected end mark"
                f" {self.end_mark!r}"
            )

        # parse name and status from first line
        tokens = self.raw_lines[0].strip().split("\t")
        name = tokens[0][2:]
        self.name = name
        status = tokens[-1]
        self.status = status
        # the rest of the lines except the last one
        lines = []
        if self.name != "Sequence Duplication Levels":
            # and column names from second/third line
            columns = self.raw_lines[1][1:].strip().split("\t")
            self._columns = columns
            for line in self.raw_lines[2:-1]:
                cols = line.strip().split("\t")
                lines.append(cols)
        else:
            extra_k, extra_v = self.raw_lines[1][1:].strip().split("\t")
            self.extra[extra_k] = convert(extra_v)
            columns = self.raw_lines[2][1:].strip().split("\t")
            self._columns = columns
            for line in self.raw_lines[3:-1]:
                cols = line.strip().split("\t")
                lines.append(cols)

        # optional processing for different modules
        if self.name == "Basic Statistics":
            return {k: convert(v) for k, v in lines}

        # try to convert numbers appropriately
        # except for "Base" column, since FastQC may output it as range
        return [
            {k: fqc_convert(k, v) for k, v in zpd}
            for zpd in [
                # zip column names and its values ~ each item in array ==
                # one row
                zip(columns, [v for v in d])
                for d in lines
            ]
        ]


class FastQC:

    """Class representing results from a FastQC run."""

    _mod_names = [
        ">>Basic Statistics",
        ">>Per base sequence quality",
        ">>Per sequence quality scores",
        ">>Per base sequence content",
        ">>Per base GC content",
        ">>Per sequence GC content",
        ">>Per base N content",
        ">>Sequence Length Distribution",
        ">>Sequence Duplication Levels",
        ">>Overrepresented sequences",
        ">>Kmer Content",
    ]

    _mod_map = {k: k.lstrip(">") for k in _mod_names}

    def __init__(
        self,
        fp: TextIO,
        max_size: int = _MAX_SIZE,
        max_line_size: int = _MAX_LINE_SIZE,
    ) -> None:
        """Initialize an instance.

        :param fp: open file handle pointing to the FastQC data file
        :param max_size: Maximum allowed size of the FastQC data file (default:
            10 MiB).
        :param max_line_size: maximum number of bytes read everytime the
            underlying ``readline`` is called (default: 1024).

        """
        self.modules = {}
        self._max_line_size = _MAX_LINE_SIZE

        line = fp.readline(self._max_line_size)
        attr = ""
        read_size = self._max_line_size
        while read_size <= max_size:

            tokens = line.strip().split("\t")
            # break on EOF
            if not line:
                break
            # parse version
            elif line.startswith("##FastQC"):
                self.version = line.strip().split()[1]
            # parse individual modules
            elif tokens[0] in self._mod_map:
                attr = self._mod_map[tokens[0]]
                raw_lines = self._read_module(fp, line)
                self.modules[attr] = FastQCModule(raw_lines)

            line = fp.readline(self._max_line_size)
            read_size += self._max_line_size

    def _read_module(self, fp: TextIO, line: str) -> List[str]:
        """Return a list of lines in a module.

        :param fp: open file handle pointing to the FastQC data file
        :param line: first line in the module
        :returns: a list of lines in the module

        """
        raw = [line]
        while not line.startswith(">>END_MODULE"):
            line = fp.readline(self._max_line_size)
            raw.append(line)

            if not line:
                raise ValueError(f"Unexpected end of file in module {line!r}")

        return raw

    @property
    def dict(self) -> Dict[str, Union[str, FastQCModulePayload]]:
        """FastQC data as a dictionary."""
        payload: Dict[str, Union[str, FastQCModulePayload]]
        payload = {k: v.dict for k, v in self.modules.items()}
        payload["version"] = self.version

        return payload


def parse(
    in_data: Union[str, PathLike, IO],
    encoding: str = "utf-8",
    results_fname: str = _RESULTS_FNAME,
    max_size: int = _MAX_SIZE,
) -> dict:
    """Parses FastQC results into a dictionary.

    :param in_data: File handle of a fastqc_data.txt file, or path to a
        fastqc_data.txt file, or path to a FastQC results directory, or path to
        a zipped FastQC result.
    :param encoding: Encoding of the input file. This is ignored if ``in_data``
        is a file handle (default: utf-8).
    :param results_fname: Name of the text file produced by FastQC in which all
        the results are stored. This is ignored if ``in_data`` is a file handle
        (default: fastqc_data.txt).
    :param max_size: Maximum allowed size of the FastQC data file (default: 10
        MiB).
    :returns: Parsed FastQC values.

    """
    # Input is zipped FastQC result, extract data file contents into a file-like
    # handle and parse it.
    if is_zipfile(in_data):
        zf = ZipFile(in_data)
        try:
            (data_fname,) = [f for f in zf.namelist() if f.endswith(results_fname)]
        except ValueError:
            raise click.BadParameter(
                f"File {in_data} contains an unexpected number of"
                f" files named {results_fname}."
            )

        with zf.open(data_fname) as src:
            data_contents = src.read(max_size).decode(encoding)

        fq = FastQC(StringIO(data_contents))

        return fq.dict

    # Make sure strings become Path instances.
    # Also, from this point on we only expect to see TextIO.
    in_data = cast(
        Union[Path, TextIO], Path(in_data) if isinstance(in_data, str) else in_data
    )

    # Input is FastQC directory.
    if isinstance(in_data, Path) and in_data.is_dir():
        try:
            ori = in_data
            in_data = ori.joinpath(next(walk(ori))[1][0], results_fname)
        except IndexError:
            raise click.BadParameter(
                f"Cannot find {results_fname} file in the given directory."
            )

    # Input is a fastqc_data.txt file handle or path to it.
    with get_handle(in_data, encoding=encoding) as fh:
        fq = FastQC(fh, max_size=max_size)

        return fq.dict



"""Parser for STAR Log.final.out file"""
# Copyright (c) 2015-2022 Wibowo Arindrarto <contact@arindrarto.dev>
# SPDX-License-Identifier: BSD-3-Clause

import os
from typing import Any, Callable, Dict, Optional, TextIO, Tuple, Union

import click

from .utils import convert, get_handle, get_linesep

__all__ = ["parse"]


def _pct_convert(raw_str: str) -> Union[str, int, float]:
    """Converts the given number ending with '%' to a number."""
    if raw_str.endswith("%"):
        return convert(raw_str[:-1])
    return convert(raw_str)


_MAX_SIZE = 1024 * 10
# Mapping of STAR attribute names to python dictionary attribute names
# and the function used to parse it.
_PARSE_MAP: Dict[str, Tuple[str, Callable[[Any], Union[str, int, float]]]] = {
    "Started job on": ("timeJobStart", str),
    "Started mapping on": ("timeMappingStart", str),
    "Finished on": ("timeEnd", str),
    "Mapping speed, Million of reads per hour": ("mappingSpeed", convert),
    "Number of input reads": ("nInput", convert),
    "Average input read length": ("avgInputLength", convert),
    "Uniquely mapped reads number": ("nUniquelyMapped", convert),
    "Uniquely mapped reads %": ("pctUniquelyMapped", _pct_convert),
    "Average mapped length": ("avgMappedLength", convert),
    "Number of splices: Total": ("nSplicesTotal", convert),
    "Number of splices: Annotated (sjdb)": ("nSplicesAnnotated", convert),
    "Number of splices: GT/AG": ("nSplicesGTAG", convert),
    "Number of splices: GC/AG": ("nSplicesGCAG", convert),
    "Number of splices: AT/AC": ("nSplicesATAC", convert),
    "Number of splices: Non-canonical": ("nSplicesNonCanonical", convert),
    "Mismatch rate per base, %": ("rateMismatchPerBase", _pct_convert),
    "Deletion rate per base": ("rateDeletionPerBase", _pct_convert),
    "Deletion average length": ("avgDeletionLength", convert),
    "Insertion rate per base": ("rateInsertionPerBase", _pct_convert),
    "Insertion average length": ("avgInsertionLength", convert),
    "Number of reads mapped to multiple loci": ("nMappedMultipleLoci", convert),
    "% of reads mapped to multiple loci": ("pctMappedMultipleLoci", _pct_convert),
    "Number of reads mapped to too many loci": ("nMappedTooManyLoci", convert),
    "% of reads mapped to too many loci": ("pctMappedTooManyLoci", _pct_convert),
    "% of reads unmapped: too many mismatches": (
        "pctUnmappedForTooManyMismatches",
        _pct_convert,
    ),
    "% of reads unmapped: too short": ("pctUnmappedForTooShort", _pct_convert),
    "% of reads unmapped: other": ("pctUnmappedForOther", _pct_convert),
}


def parse(
    in_data: Union[str, os.PathLike, TextIO],
    input_linesep: Optional[str] = None,
) -> dict:
    """Parse the log of a STAR run.

    :param in_data: Input STAR-Fusion contents.
    :param input_linesep: Name of the operating system used for determining
        input line separator. Valid values are 'nt', 'posix', or None.
    :returns: Parsed values.

    """
    payload: dict = {}
    with get_handle(in_data) as src:
        contents = src.read(_MAX_SIZE)

    linesep = get_linesep(input_linesep)
    val: Union[str, int, float]
    for line in contents.split(linesep):
        # pass empty lines
        if line and not line.strip():
            continue
        line = line.strip()
        if "|" in line:
            ori_key, val = [x.strip() for x in line.split("|", 1)]
            if ori_key in _PARSE_MAP:
                key, func = _PARSE_MAP[ori_key]
                val = func(val)
                if key in payload:
                    raise click.BadParameter(
                        f"Unexpected duplicate key entry: {key} ({ori_key})."
                    )
                payload[key] = val

    if not payload:
        msg = "Unexpected file structure. No contents parsed."
        raise click.BadParameter(msg)

    return payload



"""Parser for STAR-Fusion output"""
# Copyright (c) 2015-2022 Wibowo Arindrarto <contact@arindrarto.dev>
# SPDX-License-Identifier: BSD-3-Clause

from os import PathLike
from typing import Any, Dict, List, TextIO, Tuple, Union

import click

from .utils import get_handle

__all__ = ["parse"]

# Expected column names
# Abridged column names
_ABR_COLS = [
    "fusion_name",
    "JunctionReads",
    "SpanningFrags",
    "Splice_type",
    "LeftGene",
    "LeftBreakpoint",
    "RightGene",
    "RightBreakpoint",
]

# Non-abridged column names
_NONABR_COLS = [
    "fusion_name",
    "JunctionReads",
    "SpanningFrags",
    "Splice_type",
    "LeftGene",
    "LeftBreakpoint",
    "RightGene",
    "RightBreakpoint",
    "JunctionReads",
    "SpanningFrags",
]

# Abridged column names star-fusion 1.6.0
_ABR_COLS_v160 = [
    "FusionName",
    "JunctionReadCount",
    "SpanningFragCount",
    "SpliceType",
    "LeftGene",
    "LeftBreakpoint",
    "RightGene",
    "RightBreakpoint",
    "LargeAnchorSupport",
    "FFPM",
    "LeftBreakDinuc",
    "LeftBreakEntropy",
    "RightBreakDinuc",
    "RightBreakEntropy",
    "annots",
]

# Non-abridged column names star-fusion 1.6.0
_NONABR_COLS_v160 = [
    "FusionName",
    "JunctionReadCount",
    "SpanningFragCount",
    "SpliceType",
    "LeftGene",
    "LeftBreakpoint",
    "RightGene",
    "RightBreakpoint",
    "JunctionReads",
    "SpanningFrags",
    "LargeAnchorSupport",
    "FFPM",
    "LeftBreakDinuc",
    "LeftBreakEntropy",
    "RightBreakDinuc",
    "RightBreakEntropy",
    "annots",
]

# Abridged column names star-fusion 1.10
_ABR_COLS_v110 = [
    "FusionName",
    "JunctionReadCount",
    "SpanningFragCount",
    "est_J",
    "est_S",
    "SpliceType",
    "LeftGene",
    "LeftBreakpoint",
    "RightGene",
    "RightBreakpoint",
    "LargeAnchorSupport",
    "FFPM",
    "LeftBreakDinuc",
    "LeftBreakEntropy",
    "RightBreakDinuc",
    "RightBreakEntropy",
    "annots",
]

# Non-abridged column names star-fusion 1.10
_NONABR_COLS_v110 = [
    "FusionName",
    "JunctionReadCount",
    "SpanningFragCount",
    "est_J",
    "est_S",
    "SpliceType",
    "LeftGene",
    "LeftBreakpoint",
    "RightGene",
    "RightBreakpoint",
    "JunctionReads",
    "SpanningFrags",
    "LargeAnchorSupport",
    "FFPM",
    "LeftBreakDinuc",
    "LeftBreakEntropy",
    "RightBreakDinuc",
    "RightBreakEntropy",
    "annots",
]

# Supported columns
SUPPORTED = {
    "v1.10": _NONABR_COLS_v110,
    "v1.10_abr": _ABR_COLS_v110,
    "v1.6.0": _NONABR_COLS_v160,
    "v1.6.0_abr": _ABR_COLS_v160,
    "v0.6.0": _NONABR_COLS,
    "v0.6.0_abr": _ABR_COLS,
}

# Mapping of supported columns to output format
#
# Column name in file -> field name in json
#
# Note: the JunctionReads and SpanningFrags columns are present twice in
# the output files, with different meanings and content
# special columns are excluded
COL_MAPPING = {
    "v0.6.0": {
        "fusion_name": "fusionName",
        "JunctionReads": "nJunctionReads",
        "SpanningFrags": "nSpanningFrags",
        "Splice_type": "spliceType",
    },
    "v0.6.0_abr": {
        "fusion_name": "fusionName",
        "JunctionReads": "nJunctionReads",
        "SpanningFrags": "nSpanningFrags",
        "Splice_type": "spliceType",
    },
    "v1.6.0": {
        "FusionName": "fusionName",
        "JunctionReadCount": "nJunctionReads",
        "SpanningFragCount": "nSpanningFrags",
        "SpliceType": "spliceType",
        "JunctionReads": "junctionReads",
        "SpanningFrags": "spanningFrags",
        "LargeAnchorSupport": "largeAnchorSupport",
        "FFPM": "FFPM",
        "annots": "annots",
    },
    "v1.6.0_abr": {
        "FusionName": "fusionName",
        "JunctionReadCount": "nJunctionReads",
        "SpanningFragCount": "nSpanningFrags",
        "SpliceType": "spliceType",
        "LargeAnchorSupport": "largeAnchorSupport",
        "FFPM": "FFPM",
        "annots": "annots",
    },
    "v1.10": {
        "FusionName": "fusionName",
        "JunctionReadCount": "nJunctionReads",
        "SpanningFragCount": "nSpanningFrags",
        "est_J": "est_J",
        "est_S": "est_S",
        "SpliceType": "spliceType",
        "JunctionReads": "junctionReads",
        "SpanningFrags": "spanningFrags",
        "LargeAnchorSupport": "largeAnchorSupport",
        "FFPM": "FFPM",
        "annots": "annots",
    },
    "v1.10_abr": {
        "FusionName": "fusionName",
        "JunctionReadCount": "nJunctionReads",
        "SpanningFragCount": "nSpanningFrags",
        "est_J": "est_J",
        "est_S": "est_S",
        "SpliceType": "spliceType",
        "LargeAnchorSupport": "largeAnchorSupport",
        "FFPM": "FFPM",
        "annots": "annots",
    },
}

# Delimiter strings
_DELIM = {
    # Gene name-id
    "gids": "^",
    # Chromosome-coordinate-strand
    "loc": ":",
}


def parse_lr_entry(
    break_side: str, entries: Dict[str, str]
) -> Dict[str, Union[str, int]]:
    """Parse the gene and breakpoint entry.

    :param break_side: The side of the break, right or left
    :param entires: The entries from the current line

    """
    if break_side == "left":
        gene = entries["LeftGene"]
        breakpoint = entries["LeftBreakpoint"]
        prefix = "Left"
    elif break_side == "right":
        gene = entries["RightGene"]
        breakpoint = entries["RightBreakpoint"]
        prefix = "Right"
    else:
        raise RuntimeError("Please specify either right or left")
    gname, gid = gene.split(_DELIM["gids"])
    chrom, pos, strand = breakpoint.split(_DELIM["loc"])

    breakpoint_side = {
        "geneName": gname,
        "geneID": gid,
        "chromosome": chrom,
        "position": int(pos),
        "strand": strand,
    }  # type: Dict[str, Union[str, int]]

    # Get the other side-specific fields from the output, excluding the
    # Rigth/LeftGene and -Breakpoint fields
    sided_fields = {
        to_camel_case(field, prefix): value
        for field, value in entries.items()
        if (
            field.startswith(prefix)
            and field not in {f"{prefix}Gene", f"{prefix}Breakpoint"}
        )
    }

    breakpoint_side.update(**sided_fields)

    return breakpoint_side


def to_camel_case(field: str, prefix: str) -> str:
    """Convert a STAR-fusion output column name

    - Remove the prefix (either "Left" or "Right"
    - Convert the first character to lower case
    """
    # Remove side from field name
    new_field = field[len(prefix) :]  # noqa: E203
    # Convert to camelCase
    camel_case = new_field[0].lower() + new_field[1:]
    return camel_case


def parse_read_columns(
    colnames: List[str], values: List[str]
) -> Tuple[Dict[str, str], Dict[str, List[str]]]:
    """Parse the read columns out and return them seperately

    The JunctionReads and SpanningFrags columns can contain the actual reads,
    or just the counts of the reads. This makes parsing them correctly quite
    convoluted.

    Assumption: The last occurrence of the JunctionReads and SpanningFrags
    columns in the file contain the actual reads, while the first one contains
    the counts.
    If there is only one (as is the case for v1.6.0), it contains the reads

    :param values: List of values from the file
    :param colnames: List of column names from the file
    """
    entries = dict()
    reads = dict()

    # Keep the column names and values together
    together = [(name, val) for name, val in zip(colnames, values)]

    # Extract the JunctionReads
    rev = together[::-1]
    for name, val in rev:  # pragma: no branch
        if name == "JunctionReads":
            together.remove((name, val))
            break
    reads[name] = val.split(",")

    # Extract the SpanningFrags
    for name, val in rev:  # pragma: no branch
        if name == "SpanningFrags":
            together.remove((name, val))
            break
    reads[name] = val.split(",")

    # The other columns are regular entries
    entries = {k: v for k, v in together}

    return entries, reads


def parse_raw_line(
    raw_line: str,
    version: str,
    is_abridged: bool = True,
) -> Dict[str, Any]:
    """Parse a single line into a dictionary.

    :param raw_line: STAR-Fusion result line.
    :param version: The version of the output format present in the file.
    :param is_abridged: Whether the input raw line is from an abridged file.

    """
    values = raw_line.split("\t")
    colnames = SUPPORTED[version]

    if len(values) != len(colnames):
        msg = "Line values {0} does not match column names {1}."
        raise click.BadParameter(msg.format(values, colnames))

    if is_abridged:
        entries = {k: v for k, v in zip(colnames, values)}
        reads = dict()  # type: Dict[str, List[str]]
    else:
        # If the format is not abridged, the JunctionReads and SpanningFrags
        # columns are duplicated
        entries, reads = parse_read_columns(colnames, values)

    # Create the output dictionary based on the detected star-fusion version
    ret = dict()  # type: Dict[str, Any]
    for colname in entries:
        try:
            field_name = COL_MAPPING[version][colname]
        # For mappings that are handle elsewhere, such as LeftBreakpoint
        except KeyError:
            continue
        ret[field_name] = entries[colname]

    # Cast the apropriate entries to int
    # These values should always exist
    for int_field in ["nJunctionReads", "nSpanningFrags"]:
        ret[int_field] = int(ret[int_field])

    # Handle the special columns
    ret["left"] = parse_lr_entry("left", entries)
    ret["right"] = parse_lr_entry("right", entries)

    # Parse the annotations into a list. Not present in v0.6.0
    if "annots" in ret:
        ret["annots"] = parse_annots(ret["annots"])

    # Cast the apropriate entries to float
    # These values can be missing
    if "FFPM" in ret:
        ret["FFPM"] = float(ret["FFPM"])

    # Cast right entropy to float, not present in v0.6.0
    try:
        ret["right"]["breakEntropy"] = float(ret["right"]["breakEntropy"])
    except KeyError:
        pass

    # Cast left entropy to float, not present in v0.6.0
    try:
        ret["left"]["breakEntropy"] = float(ret["left"]["breakEntropy"])
    except KeyError:
        pass

    # Cast est_J and est_S to float, not present before v1.10
    if "est_J" in ret:
        ret["est_J"] = float(ret["est_J"])
    if "est_S" in ret:
        ret["est_S"] = float(ret["est_S"])

    if reads:
        ret["reads"] = {
            "junctionReads": reads["JunctionReads"],
            "spanningFrags": reads["SpanningFrags"],
        }
        # If there are not reads in the star-fusion output file, the column
        # will contain ".". We need to clean that up
        if ret["reads"]["junctionReads"] == ["."]:
            ret["reads"]["junctionReads"] = list()
        if ret["reads"]["spanningFrags"] == ["."]:
            ret["reads"]["spanningFrags"] = list()

    return ret


def detect_format(colnames: List[str]) -> str:
    """Return the detected column format"""
    for colformat in SUPPORTED:
        if SUPPORTED[colformat] == colnames:
            return colformat
    else:
        msg = "Unexpected column names: {0}."
        raise click.BadParameter(msg.format(colnames))


def parse_annots(annots: str) -> List[str]:
    """Split the annots field into a list"""
    # Check the format
    msg = f"Unknown annots format: {annots}"
    if not annots.startswith("[") or not annots.endswith("]"):
        raise RuntimeError(msg)

    # Cut of the square brackets
    annots = annots[1:-1]

    # Split on comma and remove quotes
    return [annotation.replace('"', "") for annotation in annots.split(",")]


def parse(in_data: Union[str, PathLike, TextIO]) -> List[dict]:
    """Parses the abridged output of a STAR-Fusion run.

    :param in_data: Input STAR-Fusion contents.

    """
    payload = []
    with get_handle(in_data) as src:
        first_line = src.readline().strip()
        if not first_line.startswith("#"):
            msg = "Unexpected header line: '{0}'."
            raise click.BadParameter(msg.format(first_line))
        # Parse column names, after removing the "#" character
        colnames = first_line[1:].split("\t")
        version = detect_format(colnames)
        is_abridged = version.endswith("_abr")
        for line in (x.strip() for x in src):
            parsed = parse_raw_line(line, version, is_abridged)
            payload.append(parsed)

    return payload



"""Parser for Picard metrics files"""
# Copyright (c) 2015-2022 Wibowo Arindrarto <contact@arindrarto.dev>
# SPDX-License-Identifier: BSD-3-Clause

import os
import re
from typing import Any, Dict, Optional, TextIO, Union

import click

from .utils import convert, get_handle, get_linesep

_MAX_SIZE = 1024 * 1024 * 1
_RE_HEADER = re.compile(r"^#+\s+")


__all__ = ["parse"]


def parse_header(header: str, linesep: str) -> Dict[str, str]:
    """Parse the Picard header lines into a dictionary.

    :param header: Raw Picard header string.
    :param linesep: Line separator characters used when parsing.

    """
    parsed = [_RE_HEADER.sub("", x) for x in header.split(linesep)]
    if len(parsed) != 4:
        raise ValueError("Unexpected Picard header.")

    return {"flags": parsed[1], "time": parsed[3]}


def parse_metrics(metrics: Optional[str], linesep: str) -> Optional[dict]:
    """Parse the Picard metrics lines into a dictionary.

    :param metris: Raw Picard metrics string.
    :param linesep: Line separator characters used when parsing.

    """
    if metrics is None:
        return None

    lines = [line.strip(linesep) for line in metrics.split(linesep)]

    metrics_class: Optional[str]
    try:
        metrics_class = lines.pop(0).split("\t")[1]
    except IndexError:
        metrics_class = None

    parsed = []
    for line in lines:
        parsed.append([convert(v) for v in line.split("\t")])

    header_cols = parsed.pop(0)
    contents: Any = [dict(zip(header_cols, line)) for line in parsed]
    if len(contents) == 1:
        contents = contents.pop()
    payload: dict = {"contents": contents}
    if metrics_class is not None:
        payload["class"] = metrics_class

    return payload


def parse_histogram(histo: Optional[str], linesep: str) -> Optional[dict]:
    """Parse the Picard histogram lines into a dictionary.

    :param metris: Raw Picard histogram string.
    :param linesep: Line separator characters used when parsing.

    """
    if histo is None:
        return None

    lines = [line.strip(linesep) for line in histo.split(linesep)]
    lines.pop(0)

    parsed = []
    for line in lines:
        parsed.append([convert(v) for v in line.split("\t")])

    header_cols = parsed.pop(0)
    payload = {"contents": [dict(zip(header_cols, line)) for line in parsed]}

    return payload


def parse(
    in_data: Union[str, os.PathLike, TextIO],
    input_linesep: Optional[str] = None,
    max_size: int = _MAX_SIZE,
) -> dict:
    """Parse an input Picard metrics file into a dictionary.

    :param in_data: Input metrics file.
    :param input_linesep: Name of the operating system used for determining
        input line separator. Valid values are 'nt', 'posix', or None.
    :param max_size: Maximum allowed size of the Picard metrics file (default:
        10 MiB).

    """
    with get_handle(in_data) as fh:
        contents = fh.read(max_size)

    linesep = get_linesep(input_linesep)
    sections = contents.strip(linesep).split(linesep * 2)

    header = next((x for x in sections if x.startswith("## htsjdk")), None)
    if header is None:
        raise click.BadParameter("Unexpected Picard metrics file format.")

    metrics = next((x for x in sections if x.startswith("## METRICS")), None)
    histo = next((x for x in sections if x.startswith("## HISTOGRAM")), None)

    return {
        "header": parse_header(header, linesep),
        "metrics": parse_metrics(metrics, linesep),
        "histogram": parse_histogram(histo, linesep),
    }



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['crimson']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'click>=8.0,<9.0',
 'single-source>=0.2.0,<0.3.0',
 'urllib3>=1.26.16,<2.0.0']

entry_points = \
{'console_scripts': ['crimson = crimson.cli:main']}

setup_kwargs = {
    'name': 'crimson',
    'version': '1.1.1',
    'description': 'Bioinformatics tool outputs converter to JSON or YAML',
    'long_description': '# Crimson\n\n[![pypi](https://img.shields.io/pypi/v/crimson)](https://pypi.org/project/crimson)\n[![ci](https://github.com/bow/crimson/actions/workflows/ci.yml/badge.svg)](https://github.com/bow/crimson/actions?query=branch%3Amaster)\n[![coverage](https://api.codeclimate.com/v1/badges/7904a5424f60f09ebbd7/test_coverage)](https://codeclimate.com/github/bow/crimson/test_coverage)\n\n\nCrimson converts non-standard bioinformatics tool outputs to JSON or YAML.\n\nCurrently it can convert outputs of the following tools:\n\n  * [FastQC](http://www.bioinformatics.babraham.ac.uk/projects/fastqc/>) (``fastqc``)\n  * [FusionCatcher](https://github.com/ndaniel/fusioncatcher) (``fusioncatcher``)\n  * [samtools](http://www.htslib.org/doc/samtools.html) flagstat (``flagstat``)\n  * [Picard](https://broadinstitute.github.io/picard/) metrics tools (``picard``)\n  * [STAR](https://github.com/alexdobin/STAR) log file (``star``)\n  * [STAR-Fusion](https://github.com/STAR-Fusion/STAR-Fusion) hits table (``star-fusion``)\n  * [Variant Effect Predictor](http://www.ensembl.org/info/docs/tools/vep/index.html)\n    plain text output (``vep``)\n\nFor each conversion, there are two execution options: as command line tool or as a Python\nlibrary function. The first alternative uses `crimson` as a command-line tool. The second one\nrequires importing the `crimson` library in your program.\n\n\n## Installation\n\nCrimson is available on the [Python Package Index](https://pypi.org/project/crimson/)\nand you can install it via ``pip``:\n\n```shell\n$ pip install crimson\n```\n\nIt is also available on\n[BioConda](https://bioconda.github.io/recipes/crimson/README.html), both through the\n`conda` package manager or as a\n[Docker container](https://quay.io/repository/biocontainers/crimson?tab=tags).\n\nFor Docker execution, you may also use\n[the GitHub Docker registry](https://github.com/bow/crimson/pkgs/container/crimson). This\nregistry hosts the latest version, but does not host versions 1.1.0 or earlier.\n\n```shell\ndocker pull ghcr.io/bow/crimson\n```\n\n\n## Usage\n\n### As a command line tool\n\nThe general command is `crimson {tool_name}`. By default, the output is written to\n`stdout`. For example, to use the `picard` parser, you would execute:\n\n```shell\n$ crimson picard /path/to/a/picard.metrics\n```\n\nYou can also write the output to a file by specifying a file name. The following\ncommand writes the output to a file named `converted.json`:\n\n```shell\n$ crimson picard /path/to/a/picard.metrics converted.json\n```\n\nSome parsers may accept additional input formats. The FastQC parser, for example, also\naccepts a path to a FastQC output directory as its input:\n\n\n```shell\n$ crimson fastqc /path/to/a/fastqc/dir\n```\n\nIt also accepts a path to a zipped result:\n\n```shell\n$ crimson fastqc /path/to/a/fastqc_result.zip\n```\n\nWhen in doubt, use the ``--help`` flag:\n\n```shell\n$ crimson --help            # for the general help\n$ crimson fastqc --help     # for the parser-specific help, in this case FastQC\n```\n\n### As a Python library function\n\nThe specific function to import is generally located at `crimson.{tool_name}.parser`. So to\nuse the `picard` parser in your program, you can do:\n\n```python\nfrom crimson import picard\n\n# You can specify the input file name as a string or path-like object...\nparsed = picard.parse("/path/to/a/picard.metrics")\n\n# ... or a file handle\nwith open("/path/to/a/picard.metrics") as src:\n    parsed = picard.parse(src)\n```\n\n## Why?\n\n  * Not enough tools use standard output formats.\n  * Writing and re-writing the same parsers across different scripts is not a productive\n    way to spend the day.\n\n\n## Local Development\n\nSetting up a local development requires that you set up all of the supported Python\nversions. We use [pyenv](https://github.com/pyenv/pyenv) for this.\n\n```shell\n# Clone the repository and cd into it.\n$ git clone https://github.com/bow/crimson\n$ cd crimson\n\n# Create your local development environment. This command also installs\n# all supported Python versions using `pyenv`.\n$ make env\n\n# Run the test and linter suite to verify the setup.\n$ make lint test\n\n# When in doubt, just run `make` without any arguments.\n$ make\n```\n\n\n## Contributing\n\nIf you are interested, Crimson accepts the following types contribution:\n\n  * Documentation updates / tweaks (if anything seems unclear, feel free to open an issue)\n  * Bug reports\n  * Support for tools\' outputs which can be converted to JSON or YAML\n\nFor any of these, feel free to open an issue in the [issue\ntracker](https://github.com/bow/crimson/issues>) or submit a pull request.\n\n\n## License\n\nCrimson is BSD-licensed. Refer to the ``LICENSE`` file for the full license.\n',
    'author': 'Wibowo Arindrarto',
    'author_email': 'contact@arindrarto.dev',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/bow/crimson',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
